from opentelemetry import propagate, trace

from pubsub_opentelemetry.subscriber_decorator import SubscriberDecorator


class TestSubscriberDecorator:
    def test_decorate(self, mocker):
        mocker.patch('opentelemetry.propagate.extract')
        propagate.extract.return_value = {'foo': 'bar'}

        message = mocker.Mock
        message.message_id = '001'
        message.attributes = {}
        message.data = b'123'

        mock_callback = mocker.Mock(return_value=message)

        def func(subscription: str, callback):
            assert subscription == 'subscription'
            return callback(message)

        tracer = trace.get_tracer(__name__)
        spy = mocker.spy(tracer, 'start_span')

        decorated = SubscriberDecorator(tracer).decorate(func)
        actual = decorated('subscription', mock_callback)
        assert actual == message
        assert mock_callback.called
        spy.assert_called_once_with(
            'subscription process',
            kind=trace.SpanKind.CONSUMER,
            attributes={
                'messaging.system': 'pubsub',
                'messaging.protocol': 'pubsub',
                'messaging.message_id': '001',
                'messaging.destination': 'subscription',
                'messaging.destination_kind': 'subscription',
                'messaging.message_payload_size_bytes': 3,
                'messaging.operation': 'process',
            },
            context={'foo': 'bar'},
        )



from typing import Callable, Collection, TypeVar

from google.cloud.pubsub_v1 import PublisherClient, SubscriberClient
from opentelemetry import trace
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor  # type: ignore
from opentelemetry.instrumentation.utils import unwrap
from wrapt import wrap_function_wrapper

from pubsub_opentelemetry.publisher_decorator import PublisherDecorator
from pubsub_opentelemetry.subscriber_decorator import SubscriberDecorator
from pubsub_opentelemetry.utils import _get_distro_version

T = TypeVar("T")


class PubSubInstrumentationProvider(BaseInstrumentor):
    """
    An instrumentor for GCP PubSub
    See `BaseInstrumentor`
    """

    @staticmethod
    def _instrument_subscribe(tracer: trace.Tracer):
        """
        Wrap the 'subscribe()' function with the instrumented decorator
        """

        def wrapper(wrapped: Callable[..., T], instance, args, kwargs):
            decorated_subscribe = SubscriberDecorator(tracer).decorate(wrapped)
            return decorated_subscribe(*args, **kwargs)

        wrap_function_wrapper(SubscriberClient, "subscribe", wrapper)

    @staticmethod
    def _instrument_publish(tracer: trace.Tracer):
        """
        Wrap the 'publish()' function with the instrumented decorator
        """

        def wrapper(wrapped: Callable[..., T], instance, args, kwargs):
            decorated_publish = PublisherDecorator(tracer).decorate(wrapped)
            return decorated_publish(*args, **kwargs)

        wrap_function_wrapper(PublisherClient, "publish", wrapper)

    def instrumentation_dependencies(self) -> Collection[str]:
        """
        The version of the instrumented library
        """
        return ['google-cloud-pubsub ~= 2.13']

    def _instrument(self, **kwargs):
        """
        Instrument the PubSub module
        """
        tracer_provider = kwargs.get("tracer_provider")
        tracer = trace.get_tracer(
            __name__,
            _get_distro_version('google-cloud-pubsub'),
            tracer_provider=tracer_provider,
        )
        self._instrument_subscribe(tracer)
        self._instrument_publish(tracer)

    def _uninstrument(self, **kwargs):
        """
        Uninstrument the PubSub module
        """
        unwrap(SubscriberClient, "subscribe")
        unwrap(PublisherClient, "publish")



import importlib_metadata

from pubsub_opentelemetry.utils import _get_distro_version


def test_get_distro_version(mocker):
    mocker.patch('importlib_metadata.version')
    importlib_metadata.version.return_value = '1.2.3'
    v = _get_distro_version('some-package')
    assert v == '1.2.3'


def test_get_not_found_distro_version(mocker):
    mocker.patch('importlib_metadata.version')
    importlib_metadata.version.side_effect = importlib_metadata.PackageNotFoundError()
    v = _get_distro_version('some-package')
    assert v == '0.0'



import importlib_metadata


def _get_distro_version(distribution: str) -> str:
    """
    Get the version of an installed distribution
    """
    try:
        version = importlib_metadata.version(distribution)
    except importlib_metadata.PackageNotFoundError:
        version = "0.0"
    return version



import wrapt
from google.cloud.pubsub_v1 import PublisherClient, SubscriberClient

from pubsub_opentelemetry import PubSubInstrumentationProvider


class TestPubSubInstrumentationProvider:
    def test_instrument(self) -> None:
        PubSubInstrumentationProvider().instrument()
        assert isinstance(PublisherClient.publish, wrapt.BoundFunctionWrapper)
        assert isinstance(SubscriberClient.subscribe, wrapt.BoundFunctionWrapper)

    def test_uninstrument(self) -> None:
        PubSubInstrumentationProvider().instrument()
        PubSubInstrumentationProvider().uninstrument()
        assert not isinstance(PublisherClient.publish, wrapt.BoundFunctionWrapper)
        assert not isinstance(SubscriberClient.subscribe, wrapt.BoundFunctionWrapper)



from typing import Any, Callable

from google.cloud.pubsub_v1.subscriber.message import Message
from opentelemetry import propagate, trace
from opentelemetry.semconv.trace import SpanAttributes


class SubscriberDecorator:
    """
    Apply the SubscriberClient decorator
    """

    def __init__(self, tracer: trace.Tracer):
        self._tracer = tracer

    def _decorate_callback(
        self, subscription: str, callback: Callable[[Message], Any]
    ) -> Callable[[Message], Any]:
        """
        Decorate the 'subscribe()' callback, start a new span
        """

        def _decorated_callback(message: Message):
            ctx = propagate.extract(message.attributes)
            attributes = {
                SpanAttributes.MESSAGING_SYSTEM: 'pubsub',
                SpanAttributes.MESSAGING_OPERATION: 'process',
                SpanAttributes.MESSAGING_DESTINATION: subscription,
                SpanAttributes.MESSAGING_DESTINATION_KIND: 'subscription',
                SpanAttributes.MESSAGING_MESSAGE_ID: message.message_id,
                SpanAttributes.MESSAGING_PROTOCOL: 'pubsub',
                SpanAttributes.MESSAGING_MESSAGE_PAYLOAD_SIZE_BYTES: len(message.data),
            }
            span = self._tracer.start_span(
                f'{subscription} process',
                kind=trace.SpanKind.CONSUMER,
                attributes=attributes,
                context=ctx,
            )
            with trace.use_span(span, end_on_exit=True):
                return callback(message)

        return _decorated_callback

    def decorate(self, subscribe: Callable) -> Callable:
        """
        Decorate the 'publish()' function, start a new span
        """

        def _decorated_subscribe(
            subscription: str, callback: Callable[[Message], Any], *args, **kwargs
        ):
            decorated_callback = self._decorate_callback(subscription, callback)
            return subscribe(subscription, decorated_callback, *args, **kwargs)

        return _decorated_subscribe



from typing import Callable

from opentelemetry import trace
from opentelemetry.semconv.trace import SpanAttributes


class PublisherDecorator:
    """
    Apply the PublisherClient decorator
    """

    def __init__(self, tracer: trace.Tracer):
        self._tracer = tracer

    def decorate(self, publish: Callable) -> Callable:
        """
        Decorate the 'publish()' function, start a new span
        """

        def _decorated_publish(topic: str, data: bytes, *args, **kwargs):
            attributes = {
                SpanAttributes.MESSAGING_SYSTEM: 'pubsub',
                SpanAttributes.MESSAGING_OPERATION: 'send',
                SpanAttributes.MESSAGING_DESTINATION: topic,
                SpanAttributes.MESSAGING_DESTINATION_KIND: 'topic',
                SpanAttributes.MESSAGING_PROTOCOL: 'pubsub',
                SpanAttributes.MESSAGING_MESSAGE_PAYLOAD_SIZE_BYTES: str(len(data)),
            }
            with self._tracer.start_as_current_span(
                f'{topic} send', kind=trace.SpanKind.PRODUCER, attributes=attributes
            ):
                return publish(topic, data, *args, **kwargs)

        return _decorated_publish



from opentelemetry import trace

from pubsub_opentelemetry.publisher_decorator import PublisherDecorator


class TestPublisherDecorator:
    def test_decorate(self, mocker) -> None:
        result = mocker.Mock
        func = mocker.Mock(return_value=result)
        tracer = trace.get_tracer(__name__)
        spy = mocker.spy(tracer, 'start_as_current_span')

        decorated = PublisherDecorator(tracer).decorate(func)
        actual = decorated('topic', b'message data')

        assert func.called
        assert actual == result
        spy.assert_called_once_with(
            'topic send',
            kind=trace.SpanKind.PRODUCER,
            attributes={
                'messaging.system': 'pubsub',
                'messaging.protocol': 'pubsub',
                'messaging.destination': 'topic',
                'messaging.destination_kind': 'topic',
                'messaging.message_payload_size_bytes': '12',
                'messaging.operation': 'send',
            },
        )



"""
This package will monkey patch the subscribe & publish functionality
within the PubSub library in order to instrument tracing.
.. _boto: https://pypi.org/project/google-cloud-pubsub/
Usage
-----
.. code:: python
    from google.cloud import pubsub_v1
    from pubsub_opentelemetry import PubSubInstrumentationProvider


    # Instrument PubSub
    PubSubInstrumentationProvider().instrument()

    # Publish creates a span with PubSub specific attributes
    with pubsub_v1.PublisherClient() as publisher:
        publisher.publish(topic_name, b'My first message!')

    # Subscribe propagates the context from the received message
    with pubsub_v1.SubscriberClient() as subscriber:
        subscriber.subscribe(subscription_name, callback)
API
---
"""

from .pubsub_instrumentation_provider import PubSubInstrumentationProvider
from .version import __version__

__all__ = ["PubSubInstrumentationProvider", "__version__"]
